# -*- coding: utf-8 -*-
#
# This file is part of INSPIRE.
# Copyright (C) 2016 CERN.
#
# INSPIRE is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# INSPIRE is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with INSPIRE; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""Configuration for crawler integration."""

from __future__ import absolute_import, print_function


CRAWLER_HOST_URL = "http://localhost:6800"
"""URL to Scrapyd HTTP server."""

CRAWLER_DATA_TYPE = "hep"
"""WorkflowObject `data_type` to set to all workflow objects."""

CRAWLER_PROJECT = "hepcrawl"
"""Scrapy project name to schedule crawls for."""

CRAWLER_SETTINGS = {
    "API_PIPELINE_URL": "http://localhost:5555/api/task/async-apply",
    "API_PIPELINE_TASK_ENDPOINT_DEFAULT": ("inspire_crawler.tasks"
                                           ".submit_results")
}
"""Dictionary of settings to add to crawlers.

By default set to flower tasks HTTP API and the standard task to be called with
the results of the harvesting.
"""

CRAWLER_SPIDER_ARGUMENTS = {}
"""Add any spider arguments to be passed when scheduling tasks.

For example for spider `myspider`:

.. code-block:: python

    {
        'myspider': {'somearg': 'foo'}
    }

You can also pass arguments directly to the scheduler with kwargs.
"""

CRAWLER_OAIHARVEST_OUTPUT_DIRECTORY = "crawls"


CRAWLER_CELERY_QUEUE = "harvests"
"""Name of the celery queue to use for harvests, if set to 'None' will not use
any specific name.
"""



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.


"""Pytest configuration."""

from __future__ import absolute_import, print_function

import os
import shutil
import tempfile

from flask import Flask
from flask.cli import ScriptInfo
from flask_celeryext import FlaskCeleryExt
from invenio_db import InvenioDB, db as db_
from invenio_workflows import InvenioWorkflows
from invenio_workflows_ui import InvenioWorkflowsUI
from invenio_oaiharvester import InvenioOAIHarvester
from inspire_crawler import INSPIRECrawler

from sqlalchemy_utils.functions import create_database, database_exists

import pytest


@pytest.yield_fixture()
def app(request):
    """Flask application fixture."""
    instance_path = tempfile.mkdtemp()
    app = Flask(__name__, instance_path=instance_path)
    app.config.update(
        CELERY_ALWAYS_EAGER=True,
        CELERY_CACHE_BACKEND="memory",
        CELERY_EAGER_PROPAGATES_EXCEPTIONS=True,
        CELERY_RESULT_BACKEND="cache",
        SECRET_KEY="CHANGE_ME",
        SECURITY_PASSWORD_SALT="CHANGE_ME_ALSO",
        SQLALCHEMY_DATABASE_URI=os.environ.get(
            'SQLALCHEMY_DATABASE_URI', 'sqlite:///test.db'),
        TESTING=True,
    )
    FlaskCeleryExt(app)
    InvenioDB(app)
    InvenioWorkflows(app)
    InvenioWorkflowsUI(app)
    InvenioOAIHarvester(app)
    INSPIRECrawler(app)

    with app.app_context():
        yield app

    shutil.rmtree(instance_path)


@pytest.yield_fixture()
def db(app):
    """Database fixture."""
    if not database_exists(str(db_.engine.url)):
        create_database(str(db_.engine.url))
    db_.create_all()
    yield db_
    db_.session.remove()
    db_.drop_all()


@pytest.fixture()
def halt_workflow(app):
    def halt_engine(obj, eng):
        return eng.halt("Test")

    class HaltTest(object):
        workflow = [halt_engine]

    app.extensions['invenio-workflows'].register_workflow(
        HaltTest.__name__, HaltTest
    )
    return HaltTest


@pytest.fixture()
def script_info(app):
    """Get ScriptInfo object for testing CLI."""
    return ScriptInfo(create_app=lambda info: app)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.


"""Module tests."""

from __future__ import absolute_import, print_function

import os
import pkg_resources
import pytest
import json
import uuid

from mock import MagicMock, PropertyMock

import requests_mock
from six.moves.urllib.parse import urlparse

from invenio_workflows import WorkflowObject, ObjectStatus
from inspire_crawler.models import JobStatus, CrawlerJob, CrawlerWorkflowObject
from inspire_crawler.tasks import submit_results
from inspire_crawler.errors import (
    CrawlerInvalidResultsPath,
    CrawlerJobNotExistError,
    CrawlerScheduleError,
    CrawlerJobError,
)
from inspire_crawler.receivers import receive_oaiharvest_job


@pytest.fixture()
def sample_records_filename():
    return pkg_resources.resource_filename(
        __name__,
        os.path.join(
            'fixtures',
            'records.jl'
        )
    )


@pytest.fixture()
def sample_records_uri(sample_records_filename):
    return "file://" + sample_records_filename


@pytest.fixture()
def sample_records(sample_records_filename):
    with open(sample_records_filename) as records_fd:
        records = (
            json.loads(line.strip()) for line in records_fd.readlines()
            if line.strip()
        )

    return list(records)


@pytest.fixture()
def sample_record_string(sample_records_filename):
    with open(sample_records_filename) as records_fd:
        line = records_fd.readline()

    return line


def test_tasks(app, db, halt_workflow, sample_records_uri):
    """Test tasks."""
    job_id = uuid.uuid4().hex  # init random value
    with app.app_context():
        with pytest.raises(CrawlerJobNotExistError):
            submit_results(
                job_id, results_uri=sample_records_uri,
                errors=None, log_file=None
            )

        CrawlerJob.create(
            job_id=job_id,
            spider="Test",
            workflow=halt_workflow.__name__,
            logs=None,
            results=None,
        )
        db.session.commit()

        with pytest.raises(CrawlerInvalidResultsPath):
            submit_results(job_id, results_uri="", errors=None, log_file=None)

    with app.app_context():
        job = CrawlerJob.get_by_job(job_id)
        assert job
        assert str(job.status)
        assert job.status == JobStatus.PENDING

        submit_results(
            job_id=job_id,
            results_uri=sample_records_uri,
            errors=None,
            log_file="/foo/bar"
        )

        job = CrawlerJob.get_by_job(job_id)
        assert job.logs == "/foo/bar"
        assert job.results == sample_records_uri

        workflow = WorkflowObject.get(1)
        assert workflow
        extra_data = workflow.extra_data
        assert 'source_data' in extra_data
        assert 'data' in extra_data['source_data']
        assert 'extra_data' in extra_data['source_data']
        expected_extra_data = {
            'crawler_job_id': job_id,
            'crawler_results_path': urlparse(sample_records_uri).path
        }
        assert expected_extra_data == extra_data['source_data']['extra_data']

        with pytest.raises(CrawlerJobError):
            submit_results(
                job_id, results_uri=sample_records_uri,
                errors=["Some error"], log_file=None
            )

        job = CrawlerJob.get_by_job(job_id)
        assert job.status == JobStatus.ERROR


def test_submit_results_with_results_data(app, db, halt_workflow,
                                          sample_records_uri, sample_records):
    """Test submit_results passing the data as payload."""
    job_id = uuid.uuid4().hex  # init random value
    with app.app_context():
        CrawlerJob.create(
            job_id=job_id,
            spider="Test",
            workflow=halt_workflow.__name__,
            logs=None,
            results=None,
        )
        db.session.commit()

    with app.app_context():
        job = CrawlerJob.get_by_job(job_id)
        assert job
        assert str(job.status)
        assert job.status == JobStatus.PENDING

        dummy_records_uri = sample_records_uri + 'idontexist'
        submit_results(
            job_id=job_id,
            results_uri=dummy_records_uri,
            results_data=sample_records,
            errors=None,
            log_file="/foo/bar"
        )

        job = CrawlerJob.get_by_job(job_id)
        assert job.logs == "/foo/bar"
        assert job.results == dummy_records_uri

        workflow = WorkflowObject.get(1)
        assert workflow
        assert workflow.extra_data['crawler_job_id'] == job_id
        crawler_results_path = workflow.extra_data['crawler_results_path']
        assert crawler_results_path == urlparse(dummy_records_uri).path

        with pytest.raises(CrawlerJobError):
            submit_results(
                job_id,
                results_uri=dummy_records_uri,
                results_data=sample_records,
                errors=["Some error"],
                log_file=None,
            )

        job = CrawlerJob.get_by_job(job_id)
        assert job.status == JobStatus.ERROR


def test_receivers(app, db, sample_record_string):
    with requests_mock.Mocker() as requests_mocker:
        job_id = uuid.uuid4().hex

        requests_mocker.register_uri(
            'POST', 'http://localhost:6800/schedule.json',
            json={'jobid': job_id, 'status': 'ok'})

        mock_record = MagicMock()
        prop_mock = PropertyMock(return_value=sample_record_string)
        type(mock_record).raw = prop_mock

        with app.app_context():
            assert receive_oaiharvest_job(
                request=None, records=[mock_record], name=""
            ) is None

            receive_oaiharvest_job(
                request=None,
                records=[mock_record],
                name='',
                spider='Test',
                workflow='test'
            )
            job = CrawlerJob.get_by_job(job_id)

            assert job


def test_receivers_exception(app, db, sample_record_string):
    with requests_mock.Mocker() as requests_mocker:
        requests_mocker.register_uri(
            'POST', 'http://localhost:6800/schedule.json',
            json={'jobid': None, 'status': 'ok'})

        mock_record = MagicMock()
        prop_mock = PropertyMock(return_value=sample_record_string)
        type(mock_record).raw = prop_mock

        with app.app_context():
            with pytest.raises(CrawlerScheduleError):
                receive_oaiharvest_job(
                    request=None,
                    records=[mock_record],
                    name='',
                    spider='Test',
                    workflow='test'
                )


def test_create_workflow_for_faulty_data(app, db, halt_workflow):
    """Test submit_results passing the data as payload."""
    job_id = uuid.uuid4().hex  # init random value
    with app.app_context():
        CrawlerJob.create(
            job_id=job_id,
            spider="desy",
            workflow=halt_workflow.__name__,
            logs=None,
            results=None,
        )
        db.session.commit()

    with app.app_context():
        job = CrawlerJob.get_by_job(job_id)
        assert job
        assert str(job.status)
        assert job.status == JobStatus.PENDING

        test_data = {
            'errors': [
                {
                    'exception': 'ValueError',
                    'traceback': 'ValueError on the line 23.'
                }
            ],
            'source_data': 'Just an XML string',
            'record': {},
            'file_name': 'broken.xml'
        }
        submit_results(
            job_id=job_id,
            results_uri='idontexist',
            results_data=[test_data],
            errors=None,
            log_file="/foo/bar"
        )
        workflow_id = CrawlerWorkflowObject.query.filter_by(job_id=job_id) \
            .one().object_id
        workflow = WorkflowObject.get(workflow_id)

        expected_crawl_error = {
            'errors': [
                {
                    'exception': 'ValueError',
                    'traceback': 'ValueError on the line 23.'
                }
            ],
            'source_data': 'Just an XML string',
            'file_name': 'broken.xml'
        }

        assert workflow.status == ObjectStatus.ERROR
        assert workflow.data == test_data['record']
        assert workflow.extra_data['crawl_errors'] == expected_crawl_error


def test_create_error_workflow_for_wrong_crawl_result(app, db, halt_workflow):
    job_id = uuid.uuid4().hex  # init random value
    with app.app_context():
        CrawlerJob.create(
            job_id=job_id,
            spider="desy",
            workflow=halt_workflow.__name__,
            logs=None,
            results=None,
        )
        db.session.commit()

    with app.app_context():
        job = CrawlerJob.get_by_job(job_id)
        assert job
        assert str(job.status)
        assert job.status == JobStatus.PENDING

        test_data = {
            'source_data': 'Just an XML string',
            'record': {},
            # missing 'errors' and 'file_name'
        }
        submit_results(
            job_id=job_id,
            results_uri='idontexist',
            results_data=[test_data],
            errors=None,
            log_file="/foo/bar"
        )
        workflow_id = CrawlerWorkflowObject.query.filter_by(job_id=job_id) \
            .one().object_id
        workflow = WorkflowObject.get(workflow_id)

        expected = {
            'errors': [
                {
                    'exception': 'KeyError',
                    'traceback': 'Wrong crawl result format. '
                                 'Missing the key `errors`'
                }
            ],
            'file_name': None,
            'source_data': {'record': {}, 'source_data': 'Just an XML string'},
        }

        assert workflow.status == ObjectStatus.ERROR
        assert workflow.data == {}
        assert workflow.extra_data['crawl_errors'] == expected



# -*- coding: utf-8 -*-
#
# This file is part of INSPIRE.
# Copyright (C) 2016 CERN.
#
# INSPIRE is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# INSPIRE is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with INSPIRE; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""Models for crawler integration."""

from __future__ import absolute_import, print_function

from datetime import datetime

from enum import Enum

from invenio_db import db

from sqlalchemy_utils.types import ChoiceType, UUIDType
from sqlalchemy.orm.exc import NoResultFound

from invenio_workflows.models import WorkflowObjectModel

from .errors import CrawlerJobNotExistError


class JobStatus(Enum):
    """Constants for possible status of any given PID."""

    __order__ = 'PENDING RUNNING FINISHED UNKNOWN'

    PENDING = 'pending'
    RUNNING = 'running'
    ERROR = 'error'
    FINISHED = 'finished'
    UNKNOWN = ''

    def __init__(self, value):
        """Hack."""

    def __eq__(self, other):
        """Equality test."""
        return self.value == other

    def __str__(self):
        """Return its value."""
        return self.value


class CrawlerJob(db.Model):
    """Keeps track of submitted crawler jobs."""

    __tablename__ = 'crawler_job'

    id = db.Column(db.Integer, primary_key=True, autoincrement=True)
    job_id = db.Column(UUIDType, index=True)
    spider = db.Column(db.String(255), index=True)
    workflow = db.Column(db.String(255), index=True)
    results = db.Column(db.Text, nullable=True)
    status = db.Column(ChoiceType(JobStatus, impl=db.String(10)),
                       nullable=False)
    logs = db.Column(db.Text, nullable=True)
    scheduled = db.Column(db.DateTime,
                          default=datetime.now,
                          nullable=False,
                          index=True)

    @classmethod
    def create(cls, job_id, spider, workflow, results=None,
               logs=None, status=JobStatus.PENDING):
        """Create a new entry for a scheduled crawler job."""
        obj = cls(
            job_id=job_id,
            spider=spider,
            workflow=workflow,
            results=results,
            logs=logs,
            status=status,
        )
        db.session.add(obj)
        return obj

    @classmethod
    def get_by_job(cls, job_id):
        """Get a row by Job UUID."""
        try:
            return cls.query.filter_by(
                job_id=job_id
            ).one()
        except NoResultFound:
            raise CrawlerJobNotExistError(job_id)

    def save(self):
        """Save object to persistent storage."""
        with db.session.begin_nested():
            db.session.add(self)


class CrawlerWorkflowObject(db.Model):
    """Relation between a job and workflow objects."""

    __tablename__ = "crawler_workflows_object"

    job_id = db.Column(UUIDType, primary_key=True)
    object_id = db.Column(
        db.Integer,
        db.ForeignKey(
            WorkflowObjectModel.id,
            ondelete="CASCADE",
            onupdate="CASCADE",
        ),
        primary_key=True
    )


__all__ = (
    'CrawlerJob',
    'CrawlerWorkflowObject',
)



# -*- coding: utf-8 -*-
#
# This file is part of INSPIRE.
# Copyright (C) 2017 CERN.
#
# INSPIRE is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# INSPIRE is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with INSPIRE; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.
"""Create inspire_crawler tables."""

from __future__ import absolute_import, print_function

import pytest
from sqlalchemy import inspect

from invenio_db.utils import drop_alembic_version_table


def test_alembic_revision_34b150f80576(app, db):
    ext = app.extensions['invenio-db']

    if db.engine.name == 'sqlite':
        raise pytest.skip('Upgrades are not supported on SQLite.')

    db.drop_all()
    drop_alembic_version_table()

    with app.app_context():
        inspector = inspect(db.engine)
        assert 'crawler_job' not in inspector.get_table_names()
        assert 'crawler_workflows_object' not in inspector.get_table_names()

    ext.alembic.upgrade(target='34b150f80576')
    with app.app_context():
        inspector = inspect(db.engine)
        assert 'crawler_job' in inspector.get_table_names()
        assert 'crawler_workflows_object' in inspector.get_table_names()

    ext.alembic.downgrade(target='df8d1c51b820')
    with app.app_context():
        inspector = inspect(db.engine)
        assert 'crawler_job' not in inspector.get_table_names()
        assert 'crawler_workflows_object' not in inspector.get_table_names()

    drop_alembic_version_table()



# -*- coding: utf-8 -*-
#
# This file is part of INSPIRE.
# Copyright (C) 2016 CERN.
#
# INSPIRE is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# INSPIRE is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with INSPIRE; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""Celery tasks for dealing with crawler."""

from __future__ import absolute_import, print_function

import copy
import json
import os

from six.moves.urllib.parse import urlparse

from celery import shared_task

from flask import current_app

from invenio_db import db

from invenio_workflows import (
    ObjectStatus,
    WorkflowEngine,
    start,
    workflow_object_class,
)

from .errors import (
    CrawlerInvalidResultsPath,
    CrawlerJobError,
    CrawlerScheduleError,
)
from .models import CrawlerJob, JobStatus, CrawlerWorkflowObject


def _extract_results_data(results_path):
    if not os.path.exists(results_path):
        raise CrawlerInvalidResultsPath(
            "Path specified in result does not exist: {0}".format(
                results_path
            )
        )

    current_app.logger.info(
        'Parsing records from {}'.format(results_path)
    )
    results_data = []
    with open(results_path) as records:
        lines = (
            line.strip() for line in records if line.strip()
        )

        for line in lines:
            current_app.logger.debug(
                'Reading line: {}'.format(line)
            )
            crawl_result = json.loads(line)
            results_data.append(crawl_result)

    current_app.logger.debug(
        'Read {} records from {}'.format(len(results_data), results_path)
    )
    return results_data


def _check_crawl_result_format(crawl_result):
    crawl_result['record']
    crawl_result['errors']
    crawl_result['source_data']
    crawl_result['file_name']


def _crawl_result_from_exception(exception, wrong_crawl_result):
    return {
        'record': {},
        'errors': [
            {
                'exception': exception.__class__.__name__,
                'traceback':
                    'Wrong crawl result format. Missing the key `{}`'
                    .format(exception.message),
            }
        ],
        'source_data': wrong_crawl_result,
        'file_name': wrong_crawl_result.get('file_name')
    }


@shared_task(ignore_results=True)
def submit_results(job_id, errors, log_file, results_uri, results_data=None):
    """Receive the submission of the results of a crawl job.

    Then it spawns the appropiate workflow according to whichever workflow
    the crawl job specifies.

    :param job_id: Id of the crawler job.
    :param errors: Errors that happened, if any (seems ambiguous)
    :param log_file: Path to the log file of the crawler job.
    :param results_uri: URI to the file containing the results of the crawl
       job, namely the records extracted.
    :param results_data: Optional data payload with the results list, to skip
        retrieving them from the `results_uri`, useful for slow or unreliable
        storages.
    """
    results_path = urlparse(results_uri).path
    job = CrawlerJob.get_by_job(job_id)
    job.logs = log_file
    job.results = results_uri

    if errors:
        job.status = JobStatus.ERROR
        job.save()
        db.session.commit()
        raise CrawlerJobError(str(errors))

    if results_data is None:
        results_data = _extract_results_data(results_path)

    for crawl_result in results_data:
        crawl_result = copy.deepcopy(crawl_result)
        try:
            _check_crawl_result_format(crawl_result)
        except KeyError as e:
            crawl_result = _crawl_result_from_exception(e, crawl_result)

        record = crawl_result.pop('record')
        crawl_errors = crawl_result['errors']

        current_app.logger.debug('Parsing record: {}'.format(record))
        engine = WorkflowEngine.with_name(job.workflow)
        engine.save()
        obj = workflow_object_class.create(data=record)
        obj.id_workflow = str(engine.uuid)
        if crawl_errors:
            obj.status = ObjectStatus.ERROR
            obj.extra_data['crawl_errors'] = crawl_result

        else:
            extra_data = {
                'crawler_job_id': job_id,
                'crawler_results_path': results_path,
            }
            record_extra = record.pop('extra_data', {})
            if record_extra:
                extra_data['record_extra'] = record_extra

            obj.extra_data['source_data'] = {
                'data': copy.deepcopy(record),
                'extra_data': copy.deepcopy(extra_data),
            }
            obj.extra_data.update(extra_data)

        obj.data_type = current_app.config['CRAWLER_DATA_TYPE']
        obj.save()
        db.session.commit()

        crawler_object = CrawlerWorkflowObject(
            job_id=job_id, object_id=obj.id
        )
        db.session.add(crawler_object)
        queue = current_app.config['CRAWLER_CELERY_QUEUE']

        if not crawl_errors:
            start.apply_async(
                kwargs={
                    'workflow_name': job.workflow,
                    'object_id': obj.id,
                },
                queue=queue,
            )

    current_app.logger.info('Parsed {} records.'.format(len(results_data)))

    job.status = JobStatus.FINISHED
    job.save()
    db.session.commit()


@shared_task(ignore_results=True)
def schedule_crawl(spider, workflow, **kwargs):
    """Schedule a crawl using configuration from the workflow objects."""
    from inspire_crawler.utils import get_crawler_instance

    crawler = get_crawler_instance()
    crawler_settings = current_app.config.get('CRAWLER_SETTINGS')
    crawler_settings.update(kwargs.get("crawler_settings", {}))

    crawler_arguments = kwargs
    crawler_arguments.update(
        current_app.config.get('CRAWLER_SPIDER_ARGUMENTS', {}).get(spider, {})
    )
    job_id = crawler.schedule(
        project=current_app.config.get('CRAWLER_PROJECT'),
        spider=spider,
        settings=crawler_settings,
        **crawler_arguments
    )
    if job_id:
        crawler_job = CrawlerJob.create(
            job_id=job_id,
            spider=spider,
            workflow=workflow,
        )
        db.session.commit()
        current_app.logger.info(
            "Scheduled scrapyd job with id: {0}".format(job_id)
        )
        current_app.logger.info(
            "Created crawler job with id:{0}".format(crawler_job.id)
        )
    else:
        raise CrawlerScheduleError(
            "Could not schedule '{0}' spider for project '{1}'".format(
                spider, current_app.config.get('CRAWLER_PROJECT')
            )
        )

    return str(crawler_job.job_id)



# -*- coding: utf-8 -*-
#
# This file is part of INSPIRE.
# Copyright (C) 2016 CERN.
#
# INSPIRE is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# INSPIRE is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with INSPIRE; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""Utility functions for interfacing with the crawler."""

from __future__ import absolute_import, print_function

from flask import current_app
from scrapyd_api import ScrapydAPI


def get_crawler_instance(*args, **kwargs):
    """Return current search client."""
    return ScrapydAPI(
        current_app.config.get('CRAWLER_HOST_URL'),
        *args,
        **kwargs
    )


def list_spiders():
    """Show the list of currently available spiders in the scrapyd server.
    """
    crawler = get_crawler_instance()
    return crawler.list_spiders(
        project=current_app.config.get('CRAWLER_PROJECT'),
    )



# -*- coding: utf-8 -*-
#
# This file is part of INSPIRE.
# Copyright (C) 2018 CERN.
#
# INSPIRE is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# INSPIRE is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with INSPIRE; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

from __future__ import absolute_import, print_function

import json

import requests_mock
from click.testing import CliRunner
from mock import patch

from inspire_crawler.cli import crawler


@patch('inspire_crawler.cli.schedule_crawl')
def test_schedule_crawl_cli(mock_schedule_crawl, script_info):
    mock_schedule_crawl.return_value = '1dd85701-787a-433d-b23f-ea4a16c5b1c0'

    runner = CliRunner()

    result = runner.invoke(
        crawler, ['schedule', 'APS', 'article'], obj=script_info)

    assert 0 == result.exit_code
    assert '1dd85701-787a-433d-b23f-ea4a16c5b1c0' in result.output


def test_list_spiders_cli(script_info):
    with requests_mock.Mocker() as requests_mocker:
        requests_mocker.register_uri(
            'GET', 'http://localhost:6800/listspiders.json?project=hepcrawl',
            json={'spiders': ['APS', 'BASE', 'CDS'], 'status': 'ok'})

        runner = CliRunner()

        result = runner.invoke(crawler, ['list-spiders'], obj=script_info)

        assert result.exit_code == 0
        assert 'APS\nBASE\nCDS\n' == result.output



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

"""Perform crawler operations such as querying crawler jobs."""

from __future__ import absolute_import, print_function

import os
import sys

import click
from flask.cli import with_appcontext
from invenio_db import db
from scrapyd_api.exceptions import ScrapydResponseError

from . import models
from .tasks import schedule_crawl
from .utils import list_spiders


CRAWLER_JOB_PROPS_TO_IGNORE = [
    'metadata',
    'query',
]


def _show_table(results):
    cols = []
    for result in results:
        if not cols:
            cols = _extract_cols_from_instance(result)
            click.secho('    '.join(cols), fg='blue')

        values = [
            str(getattr(result, col))
            for col in cols
        ]
        click.echo('    '.join(values))


def _show_file(file_path, header_name='Shown'):
    if file_path.startswith('file:/'):
        file_path = file_path[6:]

    click.secho("%s file: %s" % (header_name, file_path), fg='blue')
    if not os.path.exists(file_path):
        click.secho('    The file does not exist', fg='yellow')
    else:
        with open(file_path) as fd:
            click.echo_via_pager(fd.read())


@click.group()
def crawler():
    """Crawler commands."""


def _extract_cols_from_instance(instance):
    cols = [
        prop for prop in dir(instance)
        if not prop.startswith('_') and
        not callable(getattr(instance, prop)) and
        prop not in CRAWLER_JOB_PROPS_TO_IGNORE

    ]
    return cols


@crawler.group()
def job():
    """CrawlerJob commands."""


@job.command('list')
@click.option('--tail', default=50, help='Number of entries to show.')
@with_appcontext
def list_jobs(tail):
    """Show info about the existing crawler jobs."""
    query = (
        db.session.query(models.CrawlerJob)
        .order_by(models.CrawlerJob.id.desc())
    )
    if tail != 0:
        query = query.limit(tail)

    results = query.yield_per(10).all()
    _show_table(results=results)


@job.command('logs')
@click.argument('id')
@with_appcontext
def get_job_logs(id):
    """Get the crawl logs from the job."""
    crawler_job = models.CrawlerJob.query.filter_by(id=id).one_or_none()
    if crawler_job is None:
        click.secho(
            (
                "CrawlJob %s was not found, maybe it's not a crawl job?" %
                id
            ),
            fg='yellow',
        )
        sys.exit(1)

    if crawler_job.logs is None:
        click.secho(
            (
                "CrawlJob %s has no log, it might be that it has not run "
                "yet, you can try again later." %
                id
            ),
            fg='yellow',
        )
        sys.exit(1)

    _show_file(
        file_path=crawler_job.logs,
        header_name='Log',
    )


@job.command('results')
@click.argument('id')
@with_appcontext
def get_job_results(id):
    """Get the crawl results from the job."""
    crawler_job = models.CrawlerJob.query.filter_by(id=id).one()
    _show_file(
        file_path=crawler_job.results,
        header_name='Results',
    )


@crawler.group()
def workflow():
    """CrawlerWorkflowObject commands."""


@workflow.command('list')
@click.option('--tail', default=50, help='Number of entries to show.')
@with_appcontext
def list_crawler_workflows(tail):
    """Show info about the existing crawler workflows."""
    query = (
        models.CrawlerWorkflowObject.query
        .order_by(models.CrawlerWorkflowObject.object_id.desc())
    )
    if tail != 0:
        query = query.limit(tail)

    workflows = query.yield_per(10).all()
    _show_table(results=workflows)


@workflow.command('get_job_logs')
@click.argument('workflow_id')
@with_appcontext
def get_job_logs_from_workflow(workflow_id):
    """Retrieve the crawl logs from the workflow id."""
    query_result = (
        db.session.query(
            models.CrawlerJob.logs,
        )
        .join(
            models.CrawlerWorkflowObject,
            models.CrawlerJob.job_id == models.CrawlerWorkflowObject.job_id,
        )
        .filter(models.CrawlerWorkflowObject.object_id == workflow_id)
        .one_or_none()
    )

    if query_result is None:
        click.secho(
            (
                "Workflow %s was not found, maybe it's not a crawl workflow?" %
                workflow_id
            ),
            fg='yellow',
        )
        sys.exit(1)

    _show_file(
        file_path=query_result[0],
        header_name='Log',
    )


@workflow.command('get_job_results')
@click.argument('workflow_id')
@with_appcontext
def get_job_results_from_workflow(workflow_id):
    """Retrieve the crawl results from the workflow id."""
    query_result = (
        db.session.query(
            models.CrawlerJob.results,
        )
        .join(
            models.CrawlerWorkflowObject,
            models.CrawlerJob.job_id == models.CrawlerWorkflowObject.job_id,
        )
        .filter(models.CrawlerWorkflowObject.object_id == workflow_id)
        .one_or_none()
    )

    if query_result is None:
        click.secho(
            (
                "Workflow %s was not found, maybe it's not a crawl workflow?" %
                workflow_id
            ),
            fg='yellow',
        )
        sys.exit(1)

    _show_file(
        file_path=query_result[0],
        header_name='Results',
    )


@crawler.command('schedule')
@click.argument('spider_name')
@click.argument('workflow_name')
@click.option(
    '--dont-force-crawl',
    is_flag=True,
    help=(
        'If it should use the crawl-once mechanisms and crawl the item only '
        'when it was not crawled before.'
    ),
)
@click.option('--kwarg', multiple=True)
@with_appcontext
def schedule_crawl_cli(spider_name, workflow_name, dont_force_crawl, kwarg):
    """Schedule a new crawl.

    Note:
        Currently the oaiharvesting is done on inspire side, before this, so
        it's not supported here yet.
    """
    extra_kwargs = {}
    for extra_kwarg in kwarg:
        if '=' not in extra_kwarg:
            raise TypeError(
                'Bad formatted kwarg (%s), it should be in the form:\n'
                '    --kwarg key=value' % extra_kwarg
            )
        key, value = extra_kwarg.split('=', 1)

        extra_kwargs[key] = value

    settings = {'CRAWL_ONCE_ENABLED': False}
    if dont_force_crawl:
        settings = {}

    try:
        crawler_job_id = schedule_crawl(
            spider=spider_name,
            workflow=workflow_name,
            crawler_settings=settings,
            **extra_kwargs
        )
    except ScrapydResponseError as error:
        message = str(error)
        if 'spider' in message and 'not found' in message:
            click.echo('%s' % error)
            click.echo('\n Available spiders:')
            spiders = list_spiders()
            click.echo('\n'.join(spiders))
            raise click.Abort()
        else:
            raise

    click.echo(
        'Once the job is started, you can see the logs of the job with the '
        'command:\n'
        '    inspirehep crawler job list\n'
        '    inspirehep crawler job logs %s\n'
        '\n'
        'and for the associated workflow (it\'s job_id should be %s):\n'
        '    inspirehep crawler workflow list\n'
        % (crawler_job_id, crawler_job_id)
    )


@crawler.command('list-spiders')
@with_appcontext
def list_spiders_cli():
    """Show the list of currently available spiders in the scrapyd server.
    """
    spiders = list_spiders()
    click.echo('\n'.join(spiders))



# -*- coding: utf-8 -*-
#
# This file is part of INSPIRE.
# Copyright (C) 2016 CERN.
#
# INSPIRE is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# INSPIRE is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with INSPIRE; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization
# or submit itself to any jurisdiction.

"""Configuration for crawler integration."""

from __future__ import absolute_import, print_function

import pathlib2

from flask import current_app

from invenio_oaiharvester.signals import oaiharvest_finished
from invenio_oaiharvester.utils import write_to_dir

from .tasks import schedule_crawl


@oaiharvest_finished.connect
def receive_oaiharvest_job(request, records, name, **kwargs):
    """Receive a list of harvested OAI-PMH records and schedule crawls."""
    spider = kwargs.get('spider')
    workflow = kwargs.get('workflow')
    if not spider or not workflow:
        return

    files_created, _ = write_to_dir(
        records,
        output_dir=current_app.config['CRAWLER_OAIHARVEST_OUTPUT_DIRECTORY']
    )

    for source_file in files_created:
        # URI is required by scrapy.
        file_uri = pathlib2.Path(source_file).as_uri()
        schedule_crawl(spider, workflow, source_file=file_uri)


__all__ = (
    'receive_oaiharvest_job',
)
